#
# TRESPASSER VIRTUAL TABLE FUNCTION LABELER AND ANALYZER
#
# REQUIRES:
# - Python 2.6.x (python.exe should be added to Windows PATH)
# - PEFile package (pefile-1.2.10-63.zip or greater)
# - UNMANGLER.DLL
# - MSVCRT.DLL from Trespasser or equivalent (often already in system folder)
# - The Trespasser P611 executable (will work with others)
#
# ONLY WORKS IN WINDOWS.
#

programintro = """
Trespasser Virtual Table Function Labeler and Analyzer
v1.0
"""

programusage ="""Usage: vtflabeler executable outputfile

    executable
        Trespasser EXE file to analyze. Should be Trespasser's P6 1.1-patched
        1.0.117.0 executable or ATX's P611 executable (same). Will produce
        labels for any EXE, but the label format for the OllyDbg plugin
        will only be compatible for the P611 executable.
        
    outputfile
        Result file containing labels. May be placed in the LABELINCLUDE
        folder of the TresVTableIDLabels OllyDbg plugin; the plugin
        functions may then be executed to import the labels from this
        file into the disassembly as names.

WARNING: DOES NOT TAKE INTO ACCOUNT MULTIPLE INHERITANCE. Classes
that used multiple inheritance or that have multiple VTables will
either have erroneous output values or not be outputted at all. This
should not affect inheritance trees that don't have multiple inheritance
or such complications (which is most classes).
"""


outfiletemplate = ''';======================================================= 
; Virtual Function Labels
;======================================================= 
;
; %(numlabels)d labeled addresses
;
; These were automatically generated by a script and may contain errors.
;

'''

labeltemplate = "#label I%(classname)s::+0x%(vtoffset).3X()_virtual%(functype)s = 0x%(funcaddr).8X"
placeholdertemplate = "#label PlaceholderVirtualFunction_%(numrefs)d_VTableRefs = 0x%(funcaddr).8X"



import sys
import binascii
import struct

import pefile
import ctypes
UnmangleNameProc = ctypes.windll.unmangler.UnmangleName
UnmangleNameProc.argtypes = [ctypes.c_char_p]
UnmangleNameProc.restype = ctypes.c_char_p
FreeNameProc = ctypes.windll.unmangler.FreeName
FreeNameProc.argtypes = [ctypes.c_char_p]

    
class EXEAccessException(Exception):
    pass

class EXESection(object):
    def __init__(self,size,rva,imagebase,data):
        self.size = size
        self.rva = rva
        self.rvaend = self.size+self.rva
        self.imagebase = imagebase
        self.addr = self.rva+imagebase
        self.addrend = self.rvaend+imagebase
        self.data = data
        
    def containsaddr(self,addr):
        return (self.addr <= addr < self.addrend)
    def containsoffset(self,offset):
        return 0 <= offset < self.size

    def addrtooffset(self,addr):
        return addr - self.imagebase - self.rva
    def offsettoaddr(self,offset):
        return offset + self.imagebase + self.rva
    
    def readptr(self,offset):
        (val,) = struct.unpack('P',self.data[offset:offset+4])                   
        return val
    def readint(self,offset):
        (val,) = struct.unpack('I',self.data[offset:offset+4])                   
        return val
    def readcstr(self,offset):
        nullterm = self.data.find("\0",offset)
        return self.data[offset:nullterm]

    def readptrataddr(self,addr):     # helper
        return self.readptr(self.addrtooffset(addr))
    def readintataddr(self,addr):
        return self.readint(self.addrtooffset(addr))
    def readcstrataddr(self,addr):
        return self.readcstr(self.addrtooffset(addr))
    

class TresEXE(object):
    
    def __init__(self,filename):

        codesecname = '.text'
        datasecname = '.data'
        rdatasecname = '.rdata'

        revdict = { codesecname : None,
                    datasecname : None,
                    rdatasecname : None }

        try:
            pe = pefile.PE(filename,fast_load=True)

            for section in pe.sections:
                n = section.Name.rstrip("\s\t\0\r\n")
                if n in revdict:
                    revdict[n] = section

            for revsecs in revdict.values():
                if revsecs == None:
                    raise EXEAccessException('Missing EXE sections')


            def makesection(section,pe):
                return EXESection(section.Misc_VirtualSize,section.VirtualAddress,
                                  pe.OPTIONAL_HEADER.ImageBase,section.get_data(section.VirtualAddress))
            def makeemptysection(section,pe):
                return EXESection(section.Misc_VirtualSize,section.VirtualAddress,
                                  pe.OPTIONAL_HEADER.ImageBase,None)
            
            self.codesec = makeemptysection(revdict[codesecname],pe)
            self.datasec = makesection(revdict[datasecname],pe)
            self.rdatasec = makesection(revdict[rdatasecname],pe)

            # self.peimage = pe.get_memory_mapped_image()
            
        except Exception, err:
            raise EXEAccessException("Inaccessible or invalid Trespasser executable file.\nError: " + str(err))



class TresInterface(object):
    
    def __init__(self,vtaddr):
        self.vtaddr = vtaddr
        self.funcaddrlist = []
        self.inheritlist = []

    def addfunc(self,addr):
        self.funcaddrlist.append(addr)

    @property
    def numfuncs(self):
        return len(self.funcaddrlist)

    def addsuper(self,name):
        self.inheritlist.append(name)
    
    @property
    def numsupers(self):
        return len(self.inheritlist)


class TresFuncType(object):
    UNKNOWN = -1
    OVERRIDDEN = 1
    INHERITED = 2
    DECLARED = 3    

class TresVTAnalysis(object):
    def __init__(self,funcaddrlist):
        self.funcaddrlist = funcaddrlist
        self.funcinfolist = [TresFuncType.UNKNOWN] * len(funcaddrlist)
        self.nextconcreteparent = None
        self.numdeclared = TresFuncType.UNKNOWN

    @property
    def numfuncs(self):
        return len(self.funcinfolist)


class TresFuncLabel(object):
    def __init__(self,classname,vtindex,functype):
        self.classname = classname
        self.vtindex = vtindex
        self.functype = functype
    

class TresVTProcessor(object):

    def __init__(self,infile):
        self.exe = TresEXE(infile)
        self.icdict = {}        # TresInterface dictionary (key = class name)
        self.vtadict = {}       # TresVTAnalysis dictionary (key = class name)
        self.funcrefdict = {}       # TresFuncLabel list dictionary (key = function address)

    def scanforinterfaces(self):
    
        vtindex = 0
        vtsec = self.exe.rdatasec
        datasec = self.exe.datasec
        codesec = self.exe.codesec
        
        while vtindex < (vtsec.size-0x28):  # 0x28 is approximate point from end of section where we can stop searching (a.s.a.p.)

            # perform checks to see if vtable exists at (vtindex+4)
            typelistptr = vtsec.readptr(vtindex)
            if vtsec.containsaddr(typelistptr):

                method0 = vtsec.readptr(vtindex+4)
                if codesec.containsaddr(method0):
                    
                    typeinfoobj = vtsec.readptrataddr(typelistptr+0x0C)
                    if datasec.containsaddr(typeinfoobj):
                        
                        typevt = datasec.readptrataddr(typeinfoobj+0x00)
                        if vtsec.containsaddr(typevt):

                            # all passed!
                            ic = TresInterface(vtsec.offsettoaddr(vtindex+4))
                            
                            mangledname = self.__getvtname(typeinfoobj)    # record class name
                            self.__addvtfuncs(ic,vtindex+4) # process VTable
                            self.__scanvtinheritance(ic,typelistptr)    # process inheritance
                            # self.__showvtstats(ic,mangledname)  # stats

                            self.icdict[mangledname] = ic
                            vtindex += ic.numfuncs*4
            vtindex += 4


        # print binascii.hexlify(exe.dword(0x401000))

    def __getvtname(self,typeinfoobj):
        return self.exe.datasec.readcstrataddr(typeinfoobj+0x08)  # type_info struct
        
    def __addvtfuncs(self,ic,vtoffset):
        methodindex = vtoffset
        while True:
            nextmethod = self.exe.rdatasec.readptr(methodindex)
            if self.exe.codesec.containsaddr(nextmethod):
                ic.addfunc(nextmethod)
                methodindex += 4
            else:
                break

    def __scanvtinheritance(self,ic,typelistptr):
        vtsec = self.exe.rdatasec
        datasec = self.exe.datasec

        subptr1 = vtsec.readptrataddr(typelistptr+0x10)
        numinherit = vtsec.readintataddr(subptr1+0x08)
        lpptrlist = vtsec.readptrataddr(subptr1+0x0C)
        
        lpptroffset = lpptrlist
        for i in range(0,numinherit):
            lpptrtype = vtsec.readptrataddr(lpptroffset)
            typeptr = vtsec.readptrataddr(lpptrtype)
            typename = datasec.readcstrataddr(typeptr+0x08)
            ic.addsuper(typename)
            lpptroffset += 4

    def __showvtstats(self,ic,mangledname):
        print ""
        print "Found VTable: " + hex(ic.vtaddr)
        print "Mangled name: " + mangledname
        print str(ic.numfuncs) + " methods"
        print str(ic.numsupers) + " names in inheritance list"


    def analyzevtfuncinheritance(self):
        for icname,ic in self.icdict.iteritems():
            vta = TresVTAnalysis(ic.funcaddrlist)

            # find next class in inheritance list that actually defined new virtual functions
            for name in ic.inheritlist[1:]: # first name is always "this" class
                if name in self.icdict:
                    vta.nextconcreteparent = name
                    break

            thisvt = ic.funcaddrlist
            if vta.nextconcreteparent == None:
                parentvt = []
            else:
                parentvt = self.icdict[vta.nextconcreteparent].funcaddrlist

            if len(parentvt) > len(thisvt):
                print "VTable error: parent's VTable is larger than child's. Omitted from analysis." +\
                    "\n  Child: " + icname +\
                    "\n  Parent: " + vta.nextconcreteparent
                continue

            # check inherited functions
            for i in range(0,len(parentvt)):
                if thisvt[i] == parentvt[i]:
                    vta.funcinfolist[i] = TresFuncType.INHERITED
                else:
                    vta.funcinfolist[i] = TresFuncType.OVERRIDDEN

            # mark newly declared virtual functions
            for i in range(len(parentvt),len(thisvt)):
                vta.funcinfolist[i] = TresFuncType.DECLARED
                
            vta.numdeclared = len(thisvt) - len(parentvt);
            self.vtadict[icname] = vta

    def buildfunctionreference(self):
        for icname,vta in self.vtadict.iteritems():
            for i in range(0,vta.numfuncs):
                if vta.funcinfolist[i] == TresFuncType.OVERRIDDEN or vta.funcinfolist[i] == TresFuncType.DECLARED:
                    funcref = TresFuncLabel(icname,i,vta.funcinfolist[i])
                    self.funcrefdict.setdefault(vta.funcaddrlist[i],[]).append(funcref)
    

    def outputlabels(self,outfile,templateheader,templatelabel,templateplaceholder):
        outf = open(outfile,'w')
        outf.write(templateheader % { 'numlabels' : len(self.funcrefdict) } )

        for funcaddr in sorted(self.funcrefdict.keys()):
            funcreflist = self.funcrefdict[funcaddr]
            if len(funcreflist) == 1:
                funcref = funcreflist[0]
                if funcref.functype == TresFuncType.OVERRIDDEN:
                    fltype = '_override'
                else:
                    fltype = ''
                unmangled = UnmangleNameProc(funcref.classname)
                uparsed = unmangled
                uparsed = uparsed.replace("class ","",1)
                uparsed = uparsed.replace(" ","_")
                
                labelstr = templatelabel % { 'classname' : uparsed, 
                                             'vtoffset' : funcref.vtindex*4,
                                             'vtindex' : funcref.vtindex,
                                             'funcaddr' : funcaddr,
                                             'functype' : fltype
                                             }
                FreeNameProc(unmangled)
            else:
                labelstr = templateplaceholder % { 'numrefs' : len(funcreflist),
                                                   'funcaddr' : funcaddr
                                                   }

            outf.write("\n" + labelstr)
                
        outf.write("\n\n")
        outf.close()




def main(infile,outfile):
    try:
        processor = TresVTProcessor(infile)
        processor.scanforinterfaces()
        processor.analyzevtfuncinheritance()
        processor.buildfunctionreference()
        processor.outputlabels(outfile,outfiletemplate,labeltemplate,placeholdertemplate)
    except EXEAccessException, err:
        print str(err)
    except Exception,err:
        print "General error.\nError: " + str(err)
    print "Labels generated. See output file (" + outfile + ")."

print programintro
if len(sys.argv) != 3:
    print programusage
else:
    main(sys.argv[1],sys.argv[2])











        
